@@ -15,7 +15,7 @@
  * and /opt/lsb/lib respectively. These headers and libraries have
  * been carefully built so that they contain only the interfaces provided by
  * the LSB. And LSB conforming application must also use be linked with
- * a special program intepreter (usually ld-lsb.so.2 but this is described
+ * a special program intepreter (usually ld-lsb.so.3 but this is described
  * in each archLSB)
  * 
  * The purpose of the lsbcc tool is to cause programs to be built against
@@ -190,7 +190,7 @@ argvadd(struct argvgroup *ag, const char *opt, char *val)
 {
 char *dashopt;
 
-if( (dashopt=(char *)malloc(strlen(opt)+1)) == NULL ) {
+if( (dashopt=(char *)malloc(strlen(opt)+2)) == NULL ) {
 	fprintf(stderr,"Unable to allocate memory for an option string\n");
 	exit(4);
 	}
@@ -242,25 +242,64 @@ for(i=0;i<ag->numargv;i++)
  * a few routines here that are used to do the special processing.
  */
 
-char *default_lsblibs[] = {
-	"c",
-	"m",
-	"pthread",
-	"pam",
-	"util",
-	"dl",
+char *lsb_corelibs[] = {
+	"c",		/* core module */
 	"crypt",
-	"ncurses",
-	"z",
-	"X11",
-	"Xt",
-	"Xext",
-	"SM",
-	"ICE",
-	"GL",
-	"stdc++",
+	"dl",
 	"gcc_s",
-	0
+	"m",
+	"ncurses",
+	"pam",
+	"pthread",
+	"rt",
+	"util",
+	"z",
+	"GL",		/* graphics module */
+	"ICE",
+	"SM",
+	"X11",
+	"Xext",
+	"Xi",
+	"Xt",
+	NULL
+};
+
+char *lsb_cpluslibs[] = {
+	"stdc++",
+	NULL
+};
+
+char *lsb_desktoplibs[] = {
+	"glib-2.0",
+	"gthread-2.0",
+	"gobject-2.0",
+	"gmodule-2.0",
+	"atk-1.0",
+	"pango-1.0",
+	"pangoxft-1.0",
+	"pangoft2-1.0",
+	"gdk-x11-2.0",
+	"gdk_pixbuf_xlib-2.0",
+	"gdk_pixbuf-2.0",
+	"gtk-x11-2.0",
+	"xml2",
+	"jpeg",
+	"png12",
+	"png",
+	"fontconfig",
+        "qt-mt",
+	NULL
+};
+
+char *lsb_desktoplibs_qt4[] = {
+	"QtCore",
+	"QtGui",
+	"QtNetwork",
+	"QtXml",
+	"QtOpenGL",
+	"QtSql",
+	"QtSvg",
+	NULL
 };
 
 void
@@ -276,6 +315,7 @@ sprintf(buf,"-l%s",val);
 for(i=0;i<lsblibs->numargv;i++) {
 	if( strcmp(lsblibs->argv[i],val) == 0 ) {
 		argvaddstring(userlibs,strdup(buf));
+		/* with auto-add pthread, this isn't really needed */
 		if( strcmp(val,"pthread") == 0 ) {
 			if( lsbcc_debug&DEBUG_LIB_CHANGES )
 				fprintf(stderr,"Appending -lpthread_nonshared to the library list\n");
@@ -314,7 +354,7 @@ find_gcc_base_dir()
 {
 FILE	*cccmd;
 char	cmd[PATH_MAX];
-char	buf[PATH_MAX];
+static char	buf[PATH_MAX];
 
 /* Set a fictitious default value */
 gccbasedir="/usr/lib/gcc-lib/i386-linux/2.95.4";
@@ -334,6 +374,76 @@ gccbasedir=dirname(buf);
 
 return;
 }
+
+/*
+ * Although the C++ ABI was supposed to be the same for gcc 3.4 and
+ * 4.0, a small change was made to the ABI to fix a problem with
+ * rethrowing exception objects with complex constructors.  In order
+ * to use gcc 4.x for building LSB-compliant applications, we have to
+ * generate code using the old ABI.
+ *
+ * The ABI changes have been collapsed into a static library that
+ * lsbcc can add to the link path and resolve those symbols.  We
+ * should link that library, therefore, to all C++ applications, but
+ * only when we're using gcc 4.x.  The following function can detect
+ * when we need the compatibility library.
+ *
+ * In any situation where there's doubt, this function will report
+ * that the compatibility library is not needed.
+ */
+
+char gccversion[16] = "\0";
+
+int
+need_gcc34_compat()
+{
+  FILE *cccmd;
+  char cmd[PATH_MAX];
+  size_t resultlen;
+
+  if (gccversion[0] == '\0') {
+    /* Ask gcc for its version, if we haven't already. */
+    sprintf(cmd, "%s -dumpversion", ccname);
+
+    if ((cccmd = popen(cmd, "r")) == NULL) {
+      fprintf(stderr, "Failed to popen \"%s\"\n", cmd);
+      return 0;
+    }
+
+    fgets(gccversion, 16, cccmd);
+    pclose(cccmd);
+
+    /* Strip the trailing newline. */
+    resultlen = strlen(gccversion);
+    if (resultlen > 0) {
+      gccversion[resultlen - 1] = '\0';
+    }
+  }
+
+  /* Figure out what we need by looking at the gcc major version
+     number. */
+  switch (gccversion[0]) {
+
+  case '\0':
+    /* Read error of some kind on the gcc pipe. */
+    fprintf(stderr, "could not read gcc version\n");
+    return 0;
+
+  case '3':
+    /* gcc 3.x */
+    return 0;
+
+  case '4':
+    /* gcc 4.x */
+    return 1;
+
+  default:
+    /* Some other value we don't recognize. */
+    fprintf(stderr, "unrecognized gcc version: \"%s\"\n", gccversion);
+    return 0;
+  }
+}
+
 /* end utility functions */
 
 /*
@@ -351,7 +461,7 @@ return;
  * to list it for "special handling".  Only do this for long options
  * that are really used...
  */
-char *optstr="cL:l:o:ESI:W::su:";
+char *optstr="cL:l:o:ESI:W::su:v";
 struct option long_options[] = {
 	{"include",required_argument,0,0},
 	{"pthread",no_argument,0,0},
@@ -368,19 +478,19 @@ struct option long_options[] = {
  */
 char *proginterpreter =
 #if   __i386__
-	"/lib/ld-lsb.so.2";
+	"/lib/ld-lsb.so.3";
 #elif __powerpc__ && !__powerpc64__
-	"/lib/ld-lsb-ppc32.so.2";
+	"/lib/ld-lsb-ppc32.so.3";
 #elif __powerpc64__
-	"/lib64/ld-lsb-ppc64.so.2";
+	"/lib64/ld-lsb-ppc64.so.3";
 #elif __ia64__
-	"/lib/ld-lsb-ia64.so.2";
+	"/lib/ld-lsb-ia64.so.3";
 #elif __s390__ && !__s390x__
-	"/lib/ld-lsb-s390.so.2";
+	"/lib/ld-lsb-s390.so.3";
 #elif __s390x__
-	"/lib64/ld-lsb-s390x.so.2";
+	"/lib64/ld-lsb-s390x.so.3";
 #elif __x86_64__
-	"/lib64/ld-lsb-x86-64.so.2";
+	"/lib64/ld-lsb-x86-64.so.3";
 #else
 	"Unknown_program_interpreter";
 #endif
@@ -407,6 +517,7 @@ int main(int argc, char *argv[])
 {
 int	c,i;
 int	option_index;
+int 	auto_pthread, desktop_product, desktop_qt4_product;
 char	progintbuf[256];
 char	tmpbuf[256];
 char	*ptr;
@@ -473,13 +584,64 @@ if( (ptr=getenv("LSBCXX_INCLUDES")) != NULL ) {
 		fprintf(stderr,"c++ include prefix set to %s\n", cxxincpath );
 	}
 
+
+if( lsbcc_debug&DEBUG_ARGUMENTS ) {
+	for(i=0;i<argc;i++)
+		fprintf(stderr,"%3.3d: %s\n", i, argv[i] );
+}
+	
+/* Determine if we are being called for C or C++ */
+if( strcmp(basename(argv[0]), "lsbc++") == 0 ) {
+	/* We are compiling C++ - set a flag to affect some things later on */
+	/*
+	fprintf(stderr,"Using C++ mode\n");
+	*/
+	lsbccmode=LSBCPLUS;
+}
+
+
 /*
  * Build the argvgroup for the "known" library names here
  * Then add to it if the environment variable is set
  */
 lsblibs=argvinit();
-for(i=0;default_lsblibs[i];i++)
-	argvaddstring(lsblibs,strdup(default_lsblibs[i]));
+for(i=0;lsb_corelibs[i]; i++)
+	argvaddstring(lsblibs, strdup(lsb_corelibs[i]));
+
+if(LSBCPLUS == lsbccmode) {
+	for(i=0;lsb_cpluslibs[i]; i++)
+		argvaddstring(lsblibs, strdup(lsb_cpluslibs[i]));
+}
+
+desktop_product = 1;
+/* remove LSB_PRODUCT env var processing for LSB 3.1
+if((ptr = getenv("LSB_PRODUCT")) != NULL) {
+	if(strcasecmp(ptr, "desktop") == 0) {
+		desktop_product = 1;
+	}
+	else if (strcasecmp(ptr, "core") != 0) {
+		fprintf(stderr, "warning: LSB_PRODUCT target '%s' not valid, "
+		"must be [core|desktop], defaulting to core\n", ptr);
+		desktop_product = 0;
+	}
+} else
+	desktop_product = 0;
+*/
+desktop_qt4_product = 0;
+if((ptr = getenv("LSB_MODULES")) != NULL) {
+	if(strcasecmp(ptr, "qt4") == 0) {
+		desktop_qt4_product = 1;
+	}
+}
+
+if(desktop_product) {
+	for(i=0;lsb_desktoplibs[i];i++)
+		argvaddstring(lsblibs, strdup(lsb_desktoplibs[i]));
+}
+if(desktop_qt4_product) {
+        for(i=0;lsb_desktoplibs_qt4[i];i++)
+                argvaddstring(lsblibs, strdup(lsb_desktoplibs_qt4[i]));
+}
 
 if( (ptr=getenv("LSBCC_SHAREDLIBS")) != NULL ) {
 	char *libarg, *lib;
@@ -493,22 +655,6 @@ if( (ptr=getenv("LSBCC_SHAREDLIBS")) != NULL ) {
 		}
 	}
 
-if( lsbcc_debug&DEBUG_ARGUMENTS )
-	for(i=0;i<argc;i++)
-		fprintf(stderr,"%3.3d: %s\n", i, argv[i] );
-	
-/* Determine if we are being called for C or C++ */
-if(strcmp(basename(argv[0]),"lsbcc") == 0 ) {
-	/* We are compiling C - nothing special to do */
-	;
-} else if( strcmp(basename(argv[0]), "lsbc++") == 0 ) {
-	/* We are compiling C++ - set a flag to affect some things later on */
-	/*
-	fprintf(stderr,"Using C++ mode\n");
-	*/
-	lsbccmode=LSBCPLUS;
-}
-
 /*
  * Determine where the GCC specific file are located.
  */
@@ -535,9 +681,8 @@ if( lsbcc_debug&DEBUG_LIB_CHANGES )
 	fprintf(stderr,"Turning off default libraries with -nodefaultlibs\n");
 argvaddstring(libpaths,"-nodefaultlibs");
 if( lsbcc_debug&DEBUG_LIB_CHANGES )
-	fprintf(stderr,"Prepending %s to be linker path\n", libpath);
-sprintf(tmpbuf, "-L%s", libpath);
-argvaddstring(libpaths,strdup(tmpbuf));
+	fprintf(stderr,"Prepending %s to the linker path\n", libpath);
+argvadd(libpaths,"L",libpath);
 if( lsbcc_debug&DEBUG_LIB_CHANGES )
 	fprintf(stderr,"Prepending %s to the linker path\n",gccbasedir);
 argvadd(syslibs,"L",gccbasedir);
@@ -552,6 +697,13 @@ argvadd(syslibs,"L",gccbasedir);
 #endif
 
 if( lsbccmode == LSBCPLUS ) {
+	if ( need_gcc34_compat() ) {
+		if( lsbcc_debug&DEBUG_LIB_CHANGES )
+			fprintf(stderr,"Appending -lgcc34compat -ldl to the library list\n");
+		argvaddstring(syslibs, "-lgcc34compat");
+		argvaddstring(syslibs, "-ldl");
+	}
+
 	if( lsbcc_debug&DEBUG_LIB_CHANGES )
 		fprintf(stderr,"Appending -lstdc++ -lgcc_s to the library list\n");
 	argvaddstring(syslibs,"-lstdc++");
@@ -577,7 +729,8 @@ if( lsbccmode == LSBCPLUS ) {
 }
 
 /* Process the options passed in */
-opterr=0;
+opterr = 0;
+auto_pthread = 1;
 
 while((c=getopt_long_only(argc,argv,optstr,long_options, &option_index))>=0 ) {
 	switch(c) {
@@ -613,6 +766,7 @@ while((c=getopt_long_only(argc,argv,optstr,long_options, &option_index))>=0 ) {
 		argvreset(proginterp);
 		argvreset(syslibs);
 		argvaddstring(options,argv[optind-1]);
+		auto_pthread = 0;	/* too noisy if not linking */
 		break;
 	case 'o':
 		if( lsbcc_debug&DEBUG_RECOGNIZED_ARGS )
@@ -644,6 +798,34 @@ while((c=getopt_long_only(argc,argv,optstr,long_options, &option_index))>=0 ) {
 			fprintf(stderr,"option: -W %s\n", optarg );
 		argvaddstring(userlibs,argv[optind-1]);
 		break;
+	case 'v':
+		/* Handle the '-v' argument specially to make sure it only
+		 * calls the compiler with the '-v' option and doesn't
+		 * try to send all the other arguments to avoid the need
+		 * of having the compiler call the linker.
+		 */
+		if( lsbcc_debug&DEBUG_RECOGNIZED_ARGS )
+			fprintf(stderr,"option: -%c\n", c);
+		options=argvinit();
+		gccargs=argvinit();
+		argvaddstring(options,"-v");
+		if( lsbccmode == LSBCPLUS ) {
+		        argvaddstring(gccargs,cxxname);
+		} else {
+		        argvaddstring(gccargs,ccname);
+		}
+		argvappend(gccargs,options);
+
+		/* ensure argument list is null terminated */
+		gccargs->argv[gccargs->numargv] = NULL;
+
+		if( lsbccmode == LSBCPLUS ) {
+		        execvp(cxxname,gccargs->argv);
+		} else {
+		        execvp(ccname,gccargs->argv);
+		}
+
+		exit (EXIT_FAILURE); /* exec must have failed! */
 	case 's':
 		/*
 		 * We must explicitly recognize '-s' to distinguish it
@@ -672,6 +854,15 @@ while((c=getopt_long_only(argc,argv,optstr,long_options, &option_index))>=0 ) {
 		printf("\n");
 	}
 }
+
+/* experipment: force pthread always */
+if (auto_pthread) {
+    if (lsbcc_debug&DEBUG_LIB_CHANGES)
+	fprintf(stderr,"Appending -lpthread -lpthread_nonshared to the library list\n");
+    argvaddstring(userlibs,"-lpthread");
+    argvaddstring(userlibs,"-lpthread_nonshared");
+}
+
 /* Gather together the non-options arguments */
 if (optind < argc) {
 	if( lsbcc_debug&DEBUG_UNRECOGNIZED_ARGS )
@@ -687,7 +878,18 @@ if (optind < argc) {
 	/* There are no input file, so don't try to link anything */
 	argvreset(proginterp);
 	argvreset(syslibs);
-	}
+}
+
+/*
+ * Check if any parameters/options are passed. 
+ * If not print an error similar to running "gcc"
+ * without any parameters/options
+ */
+if(argc <= 1) {
+    fprintf(stderr, "%s: no input files\n", basename(argv[0]));
+    exit(EXIT_FAILURE);
+}
+
 
 #ifdef LSBCC_EXTRAFLAGS_HACK
 /* Gather options passed through environment variable */
